function [p,s,m] = fuzzy_clust(varargin)
% FUZZY_CLUST - Fuzzy clustering of pairwise cluster probabilities
% Fuzzy_clust converts a matrix whose entries are the probability
% for a pair of genes to be clustered together to a matrix p of 
% soft clusters, such that p(i,k) is the probability that i belongs 
% to cluster k. 
%
% USAGE:
%  [p,s,m] = first_eigen(s,K)
%    Input:  s NxN pairwise probability matrix, K number of soft clusters
%    Output: p NxK soft cluster, probabilities, s truncated input matrix,
%            m = sum(p,2) to measure what fraction of genes has been
%            assigned
%
%  [p,s,m] = first_eigen(s,K,m)
%    Use this form to continue if initial choice of K was too small.
%    Input:  s  NxN pairwise probability matrix, K number of soft clusters
%            m output variable of first run
%    Output: p NxK soft cluster, probabilities, s truncated input matrix,
%            m = sum(p,2) to measure what fraction of genes has been
%            assigned
%
% Written by
%   Anagha Joshi - anagha.joshi@psb.ugent.be
%   Tom Michoel - tom.michoel@psb.ugent.be
%
% For more information, see our paper:
%   A. Joshi, Y. Van de Peer, T. Michoel; Analysis of a Gibbs sampler
%   method for model based clustering of gene expression data (2007),
%   submitted. 

s = varargin{1};
K = varargin{2};
if nargin==3
  m = varargin{3};
end
opteigs.disp = 0;
p=sparse(length(s),K);

switch nargin
 case 2
  % initialize m
  [v,d] = eigs(s,1,'lm',opteigs);
  % check sign of v, take into account numerical precision
  [vmax, imax] = max(abs(v));
  v = v./v(imax); % this one should have all positive elements
  if length(find(v>=0)) ~= length(v)
    disp('number of negative values:')
    length(find(v<0))
    disp('lowest value:')
    min(v)
     v(find(v<0))=0;
  end
  p(:,1) = v;
  m=p(:,1);
 case 3
  m = varargin{3};
 otherwise
  error('wrong number of input arguments');
end

for k = 2:K
  k
  [i,j,s]=find(s);
  a=sqrt(1-m);
  s = a(i).*a(j).*s;
  s = sparse(i,j,s);
  [v,d] = eigs(s,1,'lm',opteigs);
  % check sign of v, take into account numerical precision
  [vmax, imax] = max(abs(v));
  v = v./v(imax); % this one should have all positive elements
  if length(find(v>=0)) ~= length(v)
    % break computation
%    disp('number of negative values:')
%    length(find(v<0))
%    disp('lowest value:')
%    min(v)
    v(find(v<0))=0;
  end
  p(:,k) = min((1-m(imax)).*v,1-m);
  m=m+p(:,k);
end
